% categorization_no_abstraction_FIG1
% This scripts makes Figures 1B through 1E in the accompanying book chapter
% What do neurons really want? The role of semantics in cortical representations
% Gabriel Kreiman

% A couple of things need to be set up first for this script to work:
% [1] The script assumes that there is a folder called 'pics'
% [2] The pics folder should have multiple subfolders, each subfolder contains images
% [3] The subfolders are based on the following ImageNet categories
% 1. n00006484_cell | 2. n02099712_labradorn | 3. 02221083_fireant | 4. n02391049_zebra | 5. n02503517_elephant | 6. n03179701_desk | 7. n03545150_house | 8. n04285008_sportscar | 9. n04971313_rose | 10. n07868340_friedrice | 11. n09618957_face | 12. n14915184_ice
% (the initial numbers are used for reference in the code below and are not part of the folder names)
% The images are NOT provided. They can be downloaded from: 
% http://image-net.org/
% (free account required)
% [4] The extension of each image file is given by image_file_extension (by default: '.JPEG')
% [5] Image size is assumed to be 56 pixels x 56 pixels x 3. Note: the images downloaded from the above website are not at this size. Scripts are available to crop and resize the images, send us an email if you want them.
% [6] The code will test the model on additional categories that it was not trained with (Figures 1D-E). These images are stored in a separate folder called 'evaluation_categories' and follow the same formatting indicated above
% [7] The code is hard wired to assume that there are two evaluation categories, but this can easily be modified to include an arbitrary number of categories

%% Parameters
codename='categorization_no_abstraction_FIG1';
list_of_categories={'n00006484_cell';'n02099712_labrador';'n02221083_fireant';'n02391049_zebra';'n02503517_elephant';'n03179701_desk';'n03545150_house';'n04285008_sportscar';'n04971313_rose';'n07868340_friedrice';'n09618957_face';'n14915184_ice'};
% 1=cell, 2=labrador, 3=fireant, 4=zebra, 5=elephan, 6=desk, 7=house, 8=sportscar, 9=rose, 10=fried rice, 10=face, 11=ice 
training_categories=[1 2 3 8 9 12];                 % This is the list of categories used to train the model in Figure 1A
                                                    % It can easily be changed to train with different categories
training_categories_text='1_2_3_8_9_12';            % Used to name the output files
n_training_categories=length(training_categories);  % Number of categories used to train the model

%training_categories=[6 10];                        % Example used in different run with model trained to distinguish desk images from fried rice images
%training_categories_text='6_10';

evaluation_categories=[7 11];                       % Categories used in Figure 1D-E to evaluate the ability to discriminate categories that the model was not trained on
evaluation_categories_text='7_11';                  % Used to name output files

layers_style=2;                                     % Dictates the architecture of the network (used in the define_layers.m function)

image_size=56;                                      % image size in pixels
imageSize=[image_size image_size 3];                % image size 

image_file_extension='.JPEG';                       % extension for images

percent_training=0.7;                               % fraction of images used for training

output_filename_core=sprintf('%s_tr%s_sem%s_sz%d_lay%d',codename,training_categories_text,evaluation_categories_text,image_size,layers_style);

tolog=1;                                            % set to 1 to write an output log file, and 0 otherwise
tosave=1;                                           % set to 1 to save the trained network and the output figures, and 0 otherwise
toplot=1;                                           % set to 1 to make figures, and 0 otherwise
if (tolog)
    log_filename=sprintf('%s.log.txt',output_filename_core);
    fid=fopen(log_filename,'w');
else
    fid=1;
end

%% Move files into the corresponding folders
% We used the following lines to automatically move the corresponding folders for training into the right directory
copy_files;

d=dir('pics');
if (length(d)<4)
    fprintf(1,'error! I cannot find the pics directory or the pics directory does not contain at least two subfolders\n');
    return;
end

%% Create image datastore 
if (~exist('imds','var'))
    fprintf(1,'Creating image datastore...\n');
    imds=imageDatastore('pics','FileExtensions',image_file_extension,'IncludeSubfolders',1,'LabelSource','foldernames');
    a=imds.countEachLabel;
    numImages_per_category=zeros(n_training_categories,1);
    for current_training_category=1:n_training_categories
        numImages_per_category(current_training_category)=a{current_training_category,2};
        fprintf(fid,'%d\t%s\t%d\n',i,list_of_categories{training_categories(current_training_category)},numImages_per_category(current_training_category));
    end
    cumsum_numImages_per_category=cumsum(numImages_per_category);
    numImages=sum(numImages_per_category);
end

%% Divide the datastore into a training and a test set
% Use percent_training x minimum number of images per category for training
numTrainingFiles = ceil(percent_training*min(numImages_per_category));
[imdsTrain,imdsTest] = splitEachLabel(imds,numTrainingFiles,'randomize');

%% Define and train the convolutional neural network architecture
if (~exist('net','var'))
    fprintf(1,'Training network...\n');
    [layers]=define_layers(imageSize,n_training_categories,layers_style);       % Define the structure of the model in Figure 1A
    
    % Set the training options to the default settings for the stochastic gradient descent with momentum
    % maximum number of epochs at 20, start the training with an initial learning rate of 0.0001
    % note: remove the plots part if you do not want to see real time of training progress
    options = trainingOptions('sgdm', ...
        'MaxEpochs',40,...
        'InitialLearnRate',1e-4, ...
        'Verbose',false, ...
        'Plots','training-progress');
    
    % Train the network
    net = trainNetwork(imdsTrain,layers,options);
    
    if (tosave)
        save_s=sprintf('save %s_net.mat net options layers',output_filename_core);
        disp(save_s);
        eval(save_s);
    end
end

%% Test the network
% Run the trained network on the test set, which was not used to train the network, and predict the image labels
YPred = classify(net,imdsTest);
YTest = imdsTest.Labels;
accuracy = sum(YPred == YTest)/numel(YTest);
chance_level=100/n_training_categories;
fprintf(fid,'training categories accuracy = %.2f [chance = %.2f]\n',100*accuracy,chance_level);

%% Compute activation of the last layer (used for Figure 1B-C)
if (~exist('featuresAll','var'))
    fprintf(1,'Computing feature activations...\n');
    featuresTrain=activations(net,imdsTrain,'fc','OutputAs','rows');
    featuresTest=activations(net,imdsTest,'fc','OutputAs','rows');
    featuresAll=activations(net,imds,'fc','OutputAs','rows');
end

if (toplot)
    figure(1);
    %% Figure 1B
    if (n_training_categories==2)
        % If there are only 2 categories, then plot the actual fc activation values without dimensionality reduction
        hs=subplot(2,2,1);
        a=imdsTrain.countEachLabel;
        n_train1=a{1,2};
        n_train2=a{2,2};
        figure;
        hs=subplot(2,2,1);
        hp=plot(featuresTrain(1:n_train1,1),featuresTrain(1:n_train1,2),'ro');
        hold on;
        hp=plot(featuresTrain((n_train1+1):(n_train1+n_train2),1),featuresTrain((n_train1+1):(n_train1+n_train2),2),'bo');
        xlabel('Feature 1');
        ylabel('Feature 2');
        title('Training');
        
        a=imdsTest.countEachLabel;
        n_test1=a{1,2};
        n_test2=a{2,2};
        hs=subplot(2,2,2);
        hp=plot(featuresTest(1:n_test1,1),featuresTest(1:n_test1,2),'ro');
        hold on;
        hp=plot(featuresTest((n_test1+1):(n_test1+n_test2),1),featuresTest((n_test1+1):(n_test1+n_test2),2),'bo');
        xlabel('Feature 1');
        ylabel('Feature 2');
        title('Test');
    else
        % If there are more than 2 categories, use tSNE to reduce the number of dimensions to 2 for plotting
        hs=subplot(2,2,1);
        if (~exist('Ytsne','var'))
            Ytsne=tsne(featuresTest);
        end
        gscatter(Ytsne(:,1),Ytsne(:,2),imdsTest.Labels,[],[],7,'off');
        xlabel('tSNE Dim1');
        ylabel('tSNE Dim2');
        set(hs,'TickDir','Out');
        set(hs,'Box','Off');
        hs=subplot(2,2,2);
        gscatter(Ytsne(:,1),Ytsne(:,2),imdsTest.Labels,[],[],7,'on');
    end
    
    %% Figure 1C
    for j=1:n_training_categories
        hs=subplot(2,n_training_categories,n_training_categories+j);
        hp=plot(featuresAll(:,j),'ko');
        if (j==1)
            i1=1;
        else
            i1=cumsum_numImages_per_category(j-1);
        end
        i2=cumsum_numImages_per_category(j);
        hold on;
        hp=plot(i1:i2,featuresAll(i1:i2,j),'bo');
        set(hp,'Color',[0.05 0.05 0.9+j/100]);
        curr_axis=axis;
        axis([0 numImages curr_axis(3) curr_axis(4)]);
        set(hs,'TickDir','Out');
        set(hs,'Box','Off');
        xlabel('Image number');
        txt=sprintf('fc unit %d',j);
        title(txt);
        if (j==1)
            ylabel('Activation strength');
        end
        for k=1:n_training_categories
            hp=plot([cumsum_numImages_per_category(k) cumsum_numImages_per_category(k)],[curr_axis(3) curr_axis(4)],'k:');
            set(hp,'Color',[0.18 0.18 0.18]);
            set(hp,'LineWidth',2);
        end
        for k=1:n_training_categories
            if (k~=j)
                if (k==1)
                    k1=1;
                else
                    k1=cumsum_numImages_per_category(k-1);
                end
                k2=cumsum_numImages_per_category(k);
                [h,p]=ttest2(featuresAll(i1:i2,j),featuresAll(k1:k2,j));
                fprintf(fid,'j=%d\tk=%d\tp=%1.2g\tmn cat1=%.2f\tstd cat1=%.2f\tmn cat2=%.2f\tstd cat2=%.2f\n',j,k,p,mean(featuresAll(i1:i2,j)),std(featuresAll(i1:i2,j)),mean(featuresAll(k1:k2,j)),std(featuresAll(k1:k2,j)));
            end
        end
    end
    if (tosave)
        print_s=sprintf('print -depsc %s_fig1.eps',output_filename_core);
        disp(print_s);
        eval(print_s);
    end
end

% Can the model discriminate other categories that it was not trained for?
%%  Create image datastore for the evaluation categories
evaluation_category1_folder=sprintf('evaluation_categories/%s',list_of_categories{evaluation_categories(1)});
evaluation_category2_folder=sprintf('evaluation_categories/%s',list_of_categories{evaluation_categories(2)});
imds_category1=imageDatastore(evaluation_category1_folder,'FileExtensions','.JPEG','IncludeSubfolders',0,'LabelSource','foldernames');
imds_category2=imageDatastore(evaluation_category2_folder,'FileExtensions','.JPEG','IncludeSubfolders',0,'LabelSource','foldernames');

features_category1=activations(net,imds_category1,'fc','OutputAs','rows');
features_category2=activations(net,imds_category2,'fc','OutputAs','rows');

%% Prepare data to use SVM to separate these two evaluation categories
n_category1=size(features_category1,1);
n_category2=size(features_category2,1);
n_training=ceil(percent_training*min([n_category1 n_category2]));
n_test=min([n_category1 n_category2])-n_training;

aux=randperm(n_category1);
indices_training_category1=aux(1:n_training);
indices_test_category1=aux((n_training+1):(n_training+n_test));

aux=randperm(n_category2);
indices_training_category2=aux(1:n_training);
indices_test_category2=aux((n_training+1):(n_training+n_test));

indices_training=[indices_training_category1(:);indices_training_category2(:)];
indices_test=[indices_test_category1(:);indices_test_category2(:)];

training_data_category1=features_category1(indices_training_category1,:);
training_data_category2=features_category2(indices_training_category2,:);
training_data=[training_data_category1;training_data_category2];
training_labels=[ones(n_training,1);2*ones(n_training,1)];
test_data_category1=features_category1(indices_test_category1,:);
test_data_category2=features_category2(indices_test_category2,:);
test_data=[test_data_category1;test_data_category2];
test_labels=[ones(n_test,1);2*ones(n_test,1)];

%% Train and test SVM
svmstruct=svmtrain(training_data,training_labels);
predictions=svmclassify(svmstruct,test_data);
accuracy = sum(predictions == test_labels)/numel(test_labels);
fprintf(fid,'evaluation categories accuracy = %.2f [chance = 50 %%]\n',100*accuracy);

if (toplot)
    %% Figure 1D
    figure(2);
    % If there are only 2 categories, then plot the actual fc activation values without dimensionality reduction
    if (n_training_categories==2)
        hs=subplot(2,2,1);
        hold on;
        for i=1:numel(test_labels)
            current_index=indices_test(i);
            if (test_labels(i)==1)
                hp=plot(features_category1(current_index,1),features_category1(current_index,2),'ro');
                if (predictions(i)==test_labels(i))
                    set(hp,'MarkerFaceColor',[1 0 0]);
                end
            else
                hp=plot(features_category2(current_index,1),features_category2(current_index,2),'bo');
                if (predictions(i)==test_labels(i))
                    set(hp,'MarkerFaceColor',[0 0 1]);
                end
            end
        end
    else
        Y=tsne(test_data);
        subplot(2,2,1);
        gscatter(Y(:,1),Y(:,2),test_labels);
        
        hs=subplot(2,2,2);hold on;
        for i=1:numel(test_labels)
            %current_index=indices_test(i);
            if (test_labels(i)==1)
                %hp=plot(features_category1(current_index,1),features_category1(current_index,2),'ro');
                hp=plot(Y(i,1),Y(i,2),'ro');
                if (predictions(i)==test_labels(i))
                    set(hp,'MarkerFaceColor',[1 0 0]);
                end
            else
                %hp=plot(features_category2(current_index,1),features_category2(current_index,2),'bo');
                hp=plot(Y(i,1),Y(i,2),'bo');
                if (predictions(i)==test_labels(i))
                    set(hp,'MarkerFaceColor',[0 0 1]);
                end
            end
            set(hp,'MarkerSize',3);
        end
        set(hs,'TickDir','Out');
        set(hs,'Box','Off');
        xlabel('tSNE Dim1');
        ylabel('tSNE Dim2');
    end
    print_s=sprintf('print -depsc %s_fig2.eps',output_filename_core);
    disp(print_s);
    eval(print_s);
    %keyboard;
end 

if (toplot)
    %% Example single unit (Figure 1E)
    fprintf(fid,'example single unit (faces versus houses)\n');
    snr=zeros(size(test_data_category1,2),1);
    for i=1:size(test_data_category1,2)
        x=test_data_category1(:,i);
        y=test_data_category2(:,i);
        snr(i)=sum(abs(x-y)./sqrt(var(x)+var(y)));
        fprintf(fid,'i=%d\tsnr(i)=%.2f\n',i,snr(i));
    end
    [mx,idx]=max(snr);
    figure(3);
    hs=subplot(2,2,1);
    hp=plot(1:length(x),test_data_category1(:,idx),'r');
    hold on;
    hp=plot((length(x)+1):(2*length(x)),test_data_category2(:,idx),'b');
    xlabel('Image number');
    ylabel('Activation strength');
    set(hs,'TickDir','Out');
    set(hs,'Box','Off');
    aux=mean(test_data_category1(:,idx));
    hp=plot([1 length(x)],[aux aux],'r--');
    set(hp,'LineWidth',2);
    aux=mean(test_data_category2(:,idx));
    hp=plot([length(x)+1 2*length(x)],[aux aux],'b--');
    set(hp,'LineWidth',2);
    curr_axis=axis;
    axis([0 2*length(x) curr_axis(3) curr_axis(4)]);

    % Replot without normalizing activity for direct comparison with Figure 1C
    hs=subplot(2,2,2);
    mx=max([max(test_data_category1(:,idx)) max(test_data_category2(:,idx))]);
    mn=min([min(test_data_category1(:,idx)) min(test_data_category2(:,idx))]);
    a=1/(mx-mn);
    b=-a*mn;
    hp=plot(1:length(x),a*test_data_category1(:,idx)+b,'r');
    hold on;
    hp=plot((length(x)+1):(2*length(x)),a*test_data_category2(:,idx)+b,'b');
    xlabel('Image number');
    ylabel('Normalized Unit Activation');
    set(hs,'TickDir','Out');
    set(hs,'Box','Off');
    aux=mean(a*test_data_category1(:,idx)+b);
    hp=plot([1 length(x)],[aux aux],'r--');
    set(hp,'LineWidth',2);
    aux=mean(a*test_data_category2(:,idx)+b);
    hp=plot([length(x)+1 2*length(x)],[aux aux],'b--');
    set(hp,'LineWidth',2);
    axis([0 2*length(x) 0 1]);
    
    if (tosave)
        print_s=sprintf('print -depsc %s_fig3.eps',output_filename_core);
        disp(print_s);
        eval(print_s);
    end
    
    fprintf(fid,'\nStatistics for faces versus houses for each fc unit (n=%d)\n',size(test_data_category1,1));
    for j=1:n_training_categories
        [h,p]=ttest2(test_data_category1(:,j),test_data_category2(:,j));
        fprintf(fid,'j=%d\tp=%1.2g\tmn faces=%.2f\tstd faces=%.2f\tmn houses=%.2f\tstd houses=%.2f\n',j,p,mean(test_data_category1(:,j)),std(test_data_category1(:,j)),mean(test_data_category2(:,j)),std(test_data_category2(:,j)));
    end
end

if (tolog)
    fclose(fid);
end



